#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include "cond.tab.h"
#include "umpf.h"
#define KLEN 10 

struct keys {
	char* keywords;
	enum yytokentype keytoks;
};

static struct keys kwds[] = 
	{ 	{"copy", KW_COPY},
	 	{"else", KW_ELSE},
		{"if", KW_IF}, 
		{"mail", KW_MAIL}, 
		{"pipe", KW_PIPE},
		{"discard", KW_DISCARD},
		{"filter", KW_FILTER}
	};

void
read_conf(char* filename)
{
	conf = fopen(filename, "r");

	if (! conf) {
		fprintf(stderr, "Error reading config file: %m\nSaving to default mailbox %s\n", default_mailbox);
		longjmp(env, 1); 
	}
}

void __attribute__ ((noreturn)) 
die(char* msg, ...)
{
        va_list args;

        va_start(args, msg);
        vfprintf(stderr, msg, args);
        fputc('\n', stderr);
        va_end(args);
        exit(1);
}

void*
xmalloc(size_t size)
{
        void* ret;

        if (!(ret = malloc(size)))
                die("Low memory");

        return ret;
}

void*
xrealloc(void* buf, size_t size)
{
	buf = realloc(buf, size);
	
	if (!buf)
		die("Low memory");

	return buf;
}

static void __attribute__ ((noreturn)) 
parse_err(char* msg, ...)
{
        va_list args;

        va_start(args, msg);
	fprintf(stderr, "Line %d: ", line);
        vfprintf(stderr, msg, args);
        fputc('\n', stderr);
	fprintf(stderr, "Saving the email to default mailbox %s\n",
		default_mailbox);
        va_end(args);
       	longjmp(env, 1); 
}

char*
xstrdup(const char* s)
{
        void* ret;

        if (!(ret = strdup(s)))
                die("Low memory");

        return ret;
}

static char*
get_string_out(int delim)
{
	int i = 0;
	int c;
	char buf[BUFSIZE];

	while ((c = getc(conf)) != delim){
		if (c == '\\') {	
			c = getc(conf);
			if (c == EOF)
				parse_err("Unbounded string, %c missing", delim);
			buf[i++] = c;
		} else if (c == '\n' || c == EOF)
			parse_err("Unbounded string, %c missing", delim);
		else		
			buf[i++] = c;

		if (i >= BUFSIZE-1)
			parse_err("Too long string, max allowed length is %d",BUFSIZE-1);	
	}	
	buf[i] = '\0';

	return xstrdup(buf);
}

static int
is_var_id(int c)
{
	return 	(c >= '0' && c <= '9' ) ||
		(c >= 'a' && c <= 'z') ||
		(c >= 'A' && c <= 'Z') ||
		c == '_' ||
		c == '-'; 
}

static int
is_alpha(int c)
{
	return (c >= 'a' && c <= 'z') ||
                (c >= 'A' && c <= 'Z');
}

static int
get_token_start(void)
{
	int c;
	int want_newline = 0;

	for(;;) {
		c = getc(conf);
		if (c == EOF)
			return c;
		if (c == '#') {
			want_newline = 1;
			continue;
		}
		if (c == '\n') {
			line++;
			want_newline = 0;
			continue;
		}
		if (c != '\n' && want_newline)
			continue;
		if (c != ' ' && c != '\t')
			return c;
	}
}

int
yylex(void)
{
	int c;
	
	c = get_token_start();
	if (c == EOF)
		return 0;
	
	int d = getc(conf);
	if (d >= 0) {
		switch (CC(c,d)) {
		case CC('!','='): yylval.n = CC('!','='); return NEQ;
		case CC('!','~'): yylval.n = CC('!','='); return NRE;
		case CC('<','='): yylval.n = CC('<','='); return LE;
		case CC('>','='): yylval.n = CC('>','='); return GE;
		case CC('=','='): yylval.n = CC('=','='); return EQ;
		case CC('~','~'): yylval.n = CC('~','~'); return RE;
		case CC('-','>'): yylval.n = CC('-','>'); return ARROW;
		}
		ungetc(d,conf);
	}

	switch (c) {
		case '!':
		case '(':
		case ')':
		case '-':
		case '+':
		case '*':
		case '/':
		case '{':
		case '}':
		case '<':
		case '>':
		case '=':
		case ';':
		case '.':
		case '|':
		case '&':
		case '^':
			yylval.n = c;
			return c;
		
		case '"':
		case '\'':
			yylval.str = get_string_out(c);
			return CONST;	
	}

	if (c >= '0' && c <= '9'){
		ungetc(c,conf);
		int i = 0;
		char buf[BUFSIZE];
	
		while ((c = getc(conf))>= '0' && c<= '9'){
			buf[i] = c;
			i++;
			if (i >= BUFSIZE-1)
				parse_err("Too long number");	
		}
		if (c != EOF)
			ungetc(c,conf);
		buf[i] = 0;
		yylval.str = xstrdup(buf);

		return CONST;
	}

	if (c == '$'){
		int i = 0;
		char buf[BUFSIZE];
	
		while (is_var_id(c = getc(conf))){
			buf[i]=c;
			i++;
			if (i >= BUFSIZE-1)
				parse_err("Too long identifier, max allowed length is %d",BUFSIZE-1);	
		}
		if (!i)
			parse_err("Variable identifier must not be empty");
		if (c > 0)
			ungetc(c,conf);
		buf[i] = 0;
		yylval.str = xstrdup(buf);

		return VAR;
	}

	if (is_alpha(c)){
		char buf[KLEN];	
		int n, i = 0;

		ungetc(c,conf);
		while (is_alpha(c = getc(conf))){
			buf[i++] = c;
			if (i >= KLEN)
				parse_err("Keyword too long");
		}
		buf[i] = 0;
		if (c > 0)
			ungetc(c,conf);

		n = (sizeof(kwds)/sizeof(struct keys));
		for (i = 0; i < n; i++){
			if (!strcmp(buf,kwds[i].keywords))
				return kwds[i].keytoks;
		}

		parse_err("Unknown keyword %s", buf);
	}

	parse_err("Unknown character %c", c);
}
