#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <time.h>

#include "umpf.h"

#define LOCK_MAX_TRIES 3
gid_t rgid, egid, sgid;

void
save_gids(void)
{
	getresgid(&rgid, &egid, &sgid);	
}

static void
drop_gid(void)
{
	setresgid(rgid, rgid, egid);
}

static void
raise_gid(void)
{
	setresgid(rgid, sgid, sgid);
}

static void
random_sleep(unsigned int about, unsigned int offset)
{
	int myrand;

	if (about == 0 || offset > about)
		die("random sleep: Bad input data: %d +- %d", about, offset);

	myrand = random() % offset * (random()%2?1:-1);

	usleep(about * 1000000 + myrand * 500000);
}

char*
cat(char* l, char* r)
{
	char* res = xmalloc(strlen(l) + strlen (r) + 1);
	strcpy(res, l);
	strcat(res, r);
	
	return res;	
}

static int
dot_lock(char* path)
{
	int i;
	int fd;
	int res = -1;
	char* lockfile = cat(path, ".lock");

	raise_gid();
	for (i = 0; i < LOCK_MAX_TRIES; i++){
		struct stat buf;

		if ((fd = open(lockfile, O_WRONLY | O_EXCL | O_CREAT, 0)) 
			>= 0) {
			close(fd);
			res = 0;
			break;
        	}

		if (errno != EEXIST)
			break;
		/* deal with old locks */
		stat(lockfile, &buf);
		if ((time(NULL) - buf.st_mtime) > 3600)
			unlink(lockfile);		
		random_sleep(1, 1);
	}
	drop_gid();

	free(lockfile);
	return res;
}

static void
dot_unlock(char* path)
{
	char* lockfile = cat(path, ".lock");

	raise_gid();
	unlink(lockfile);
	drop_gid();
	free(lockfile);
}

static int
do_open_mailbox(char* path, int is_default_mailbox)
{
	int fd;

	/* attempt to dot_lock first */
	if (dot_lock(path))
		return -1;

	/* if OK, try to open the file:
		either we are saving to default mailbox (no problem here)
		or we are saving somewhere else (no need to be privileged)
	*/
	int perms = S_IRUSR | S_IWUSR;
	if (is_default_mailbox)
		perms |=  (S_IRGRP | S_IWGRP);

	mode_t oldmask = umask(0);
	fd = open(path, O_RDWR | O_CREAT, perms);
	umask(oldmask);
	if (fd < 0){
		dot_unlock(path);
		return -1;
	}	

	/* fcntl then */
	struct flock mb_lock;
	memset(&mb_lock, 0, sizeof(struct flock));
	mb_lock.l_type = F_WRLCK;
	mb_lock.l_whence = SEEK_SET;

	if (fcntl(fd, F_GETLK, &mb_lock) < 0){
		close(fd);
		dot_unlock(path);
		return -1;
	}

	return fd; 
}

static void
do_close_mailbox(int fd, char* path)
{
	struct flock mb_lock;
	memset(&mb_lock, 0, sizeof(struct flock));
	mb_lock.l_type = F_UNLCK; 
	fcntl(fd, F_SETLK, &mb_lock);
	close(fd);
	dot_unlock(path);
}

void
close_mailbox(int fd, char* path, int is_default_mailbox)
{
	if (is_default_mailbox)
		raise_gid();
	do_close_mailbox(fd, path);
	if (is_default_mailbox)
		drop_gid();
}

int
open_mailbox(char* path, int is_default_mailbox)
{
	int fd;

	if (is_default_mailbox)
		raise_gid();
	fd = do_open_mailbox(path, is_default_mailbox);
	if (is_default_mailbox)
		drop_gid();

	return fd;
}
